(*
 * COMS4115: Pseudo Abstract Syntax Tree
 *
 * Authors:
 *  - Raymond Xu
 *  - Kevin Lin
 *  - Kristy Choi
 *  - Dennis Wei
 *  - Benjamin Low
 *)

(* Unary operators *)
type unop =
  | Neg      (* - *)
  | Not      (* !/not *)

(* Binary operators *)
type binop =
  (* Arithmetic *)
  | Add      (* + *)
  | Minus    (* - *)
  | Times    (* * *)
  | Divide   (* / *)
  (* Boolean *)
  | Or       (* || *)
  | And      (* && *)
  | Eq       (* == *)
  | Neq      (* != *)
  | Less     (* < *)
  | Leq      (* <= *)
  | Greater  (* > *)
  | Geq      (* >= *)
  (* String *)
  | Concat   (* ^ *)
  (* List *)
  | Combine  (* :: *)

(* Expressions *)
(*type num =
  | Num_int of int      (* 42 *)
  | Num_float of float  (* 42.0 *)*)

type expr =
  | Num_lit of float               (* 42 *)
  | String_lit of string          (* "Hello, world" *)
  | Bool_lit of bool              (* true *)
  | Unop of unop * expr           (* -5 *)
  | Binop of expr * binop * expr  (* a + b *)
  | Id of string                  (* x *)
  | Assign of string * expr       (* x = 4 *)
  | Call of string * expr list      (* add(1, 2) *)
  | ListDecl of expr list         (* [1, 2, 3] *)
  | DictDecl of (expr * expr) list         
  (* List Operations *)
  | ListInsert of expr * expr * expr
  | ListPush of expr * expr
  | ListRemove of expr * expr
  | ListPop of expr
  | ListDequeue of expr
  | ListLength of expr
  | ListGet of expr * expr
  | ListSet of expr * expr * expr
  (* Dict Operations *)
  | DictMem of expr * expr
  | DictFind of expr * expr
  | DictMap of expr * expr * expr
  | DictDelete of expr * expr
  | DictSize of expr
  | Noexpr
  | ObjectField of expr * string
  | ObjectAssign of expr * string * expr

(* Statements *)
type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | Break
  | Continue
  | While of expr * stmt
  | ForIn of expr * expr * stmt
  | ForRange of expr * expr * expr * stmt
  | If of expr * stmt * stmt
  | Print of expr
  | ObjectInit of string list

(* Function Declarations *)
type func_decl = {
  fname: string;
  formals: string list;  (* Parameters *)
  body: stmt list;      (* Function Body *)
}

(* Program entry point *)
type program = func_decl list
