open Ast
open Sast
open Printf

(* Unary operators *)
let txt_of_unop = function
  | Neg -> "\"Neg\""
  | Not -> "\"Not\""

let txt_of_prim = function
  | Num -> "Num"
  | Bool -> "Bool"
  | String -> "String"
  | Void -> "Void"
  | _ -> "Not prim"

(* Data type *)
let txt_of_data_type = function
  | Num -> "\"Num\""
  | Bool -> "\"Bool\""
  | String -> "\"String\""
  | Void -> "\"Void\""
  | Object(id) -> sprintf "\"Object(%s)\"" (string_of_int id)
  | List(d) -> sprintf "\"List(%s)\"" (txt_of_prim d)
  | Undetermined -> "\"Undetermined\""
  | T(s) -> sprintf "\"T(%s)\"" s


(* Binary operators *)
let txt_of_binop = function
  (* Arithmetic *)
  | Add -> "\"Add\""
  | Minus -> "\"Minus\""
  | Times-> "\"Times\""
  | Divide -> "\"Divide\""
  (* Boolean *)
  | Or -> "\"Or\""
  | And -> "\"And\""
  | Eq -> "\"Eq\""
  | Neq -> "\"Neq\""
  | Less -> "\"Less\""
  | Leq -> "\"Leq\""
  | Greater -> "\"Greater\""
  | Geq -> "\"Geq\""
  (* List *)
  | Combine  -> "\"Combine\""
  (* String *)
  | Concat -> "\"Concat\""

(* Expressions *)
(*let txt_of_num = function
  | Num_int(x) -> string_of_int x
  | Num_float(x) -> string_of_float x *)

let rec txt_of_aexpr = function
  | ANumLit(x, t) -> sprintf "{ \"aexpr\": \"ANumLit\",     \"data_type\": %s,     \"val\" : \"%s\" } " (txt_of_data_type t) (string_of_float x)
  | ABoolLit(x, t) -> sprintf "{ \"aexpr\": \"ABoolLit\",     \"data_type\": %s,     \"val\": \"%s\" }" (txt_of_data_type t) (string_of_bool x) 
  | AStringLit(x, t) -> sprintf "{ \"aexpr\": \"AStringLit\",     \"data_type\": %s,     \"val\": \"%s\" }" (txt_of_data_type t) x
  | AVal(s, d) -> sprintf "{ \"aexpr\": \"AVal\",     \"val_name\": \"%s\",     \"data_type\": %s }" s (txt_of_data_type d)
  (*| Id(x) -> sprintf "Id(%s)" x *)
  | ABinop(e1, op, e2, t) -> sprintf "{ \"aexpr\": \"ABinop\",     \"e1\": %s,     \"op\": %s,     \"e2\": %s,     \"val\": %s }"
      (txt_of_aexpr e1) (txt_of_binop op) (txt_of_aexpr e2) (txt_of_data_type t)
  | AUnop(op, e, t) -> sprintf "{ \"aexpr\": \"AUnop\",     \"e\": %s,     \"op\": %s,     \"val\": %s }" (txt_of_unop op) (txt_of_aexpr
  e) (txt_of_data_type t)
  | AAssign(x, e, t) -> sprintf "{ \"aexpr\": \"AAssign\",     \"var\": \"%s\",     \"e\": %s,     \"data_type\": %s }" x (txt_of_aexpr e)
  (txt_of_data_type t)
  | ACall(f, args, t) -> sprintf "{ \"aexpr\": \"AACall\",     \"name\": \"%s\", \"args\": [%s],     \"data_type\": %s }"
      f (txt_of_list args) (txt_of_data_type t)
  | ANoexpr(t) -> sprintf "Noexpr(%s)" (txt_of_data_type t)

  (* List Operations *)
  | AListDecl(l, d) -> sprintf "{ \"aexpr\": \"AListDecl\",     \"val\": [%s],     \"data_type\": %s }" (txt_of_list l) (txt_of_data_type d)
  | AListInsert(e1, e2, e3, d) -> sprintf "{ \"aexpr\": \"AListInsert\",     \"e1\": %s,     \"e2\": %s,     \"e3\": %s,\"data_type\": %s }"  (txt_of_aexpr e1) (txt_of_aexpr e2) (txt_of_aexpr e3) (txt_of_data_type d)
  | AListPush(e1, e2, d) -> sprintf "{ \"aexpr\": \"AListPush\",     \"e1\": %s,     \"e2\": %s,     \"data_type\": %s }"  (txt_of_aexpr e1) (txt_of_aexpr e2) (txt_of_data_type d)
  | AListRemove(e1, e2, d) -> sprintf "{ \"aexpr\": \"AListRemove\",     \"e1\": %s,     \"e2\": %s,     \"data_type\": %s }"  (txt_of_aexpr e1) (txt_of_aexpr e2) (txt_of_data_type d)
  | AListPop(e, d) -> sprintf "{ \"aexpr\": \"AListPop\",     \"e\": %s,     \"data_type\": %s }"  (txt_of_aexpr e) (txt_of_data_type d)
  | AListDequeue(e, d) -> sprintf "{ \"aexpr\": \"AListDequeue\",     \"e\": %s,     \"data_type\": %s }"  (txt_of_aexpr e) (txt_of_data_type d)
  | AListLength(e, d) -> sprintf "{ \"aexpr\": \"AListLength\",     \"e\": %s,     \"data_type\": %s }"  (txt_of_aexpr e) (txt_of_data_type d)
  | AListGet(e1, e2, d) -> sprintf "{ \"aexpr\": \"AListRemove\",     \"e1\": %s,     \"e2\": %s,     \"data_type\": %s }"  (txt_of_aexpr e1) (txt_of_aexpr e2) (txt_of_data_type d)
  | AListSet(e1, e2, e3, d) -> sprintf "{ \"aexpr\": \"AListSet\",     \"e1\": %s,     \"e2\": %s,     \"e3\": %s,\"data_type\": %s }"  (txt_of_aexpr e1) (txt_of_aexpr e2) (txt_of_aexpr e3) (txt_of_data_type d)
  (* Objects *)
  | AObjectField(e, s, d) -> sprintf "{ \"aexpr\": \"AObjectField\",     \"e\": %s,     \"s\": \"%s\",     \"data_type\": %s }" (txt_of_aexpr e) s (txt_of_data_type d)
  | AObjectAssign(e1, s, e2, d) -> sprintf "{ \"aexpr\": \"AObjectAssign\", \"e1\": %s,     \"s\": \"%s\",     \"e2\": %s,     \"data_type\": %s }" (txt_of_aexpr e1) s (txt_of_aexpr e2) (txt_of_data_type d)

and txt_of_objlist obj_list =
  let rec aux acc = function
      | [] -> sprintf "[%s]" (String.concat " , " (acc))
      | obj :: tl -> aux (sprintf "(%s, %s)" (fst obj) (txt_of_data_type (snd obj)) :: acc) tl
  in aux [] obj_list

(* Lists *)
and txt_of_list = function
  | [] -> ""
  | [x] -> txt_of_aexpr x
  | _ as l -> String.concat " , " (List.map txt_of_aexpr l)

and txt_of_aformal = function
    | (d, s) -> sprintf "{ %s:\"%s\" }" (txt_of_data_type d) s

and txt_of_aformals (aformals) =
  let rec aux acc = function
      | [] -> sprintf "%s" (String.concat " , " (List.rev acc))
      | aformal :: tl -> aux (txt_of_aformal aformal:: acc) tl
  in aux [] aformals 

(* Statements *)
and txt_of_astmt = function
    | AExpr(e) -> sprintf "  { \"astmt\": \"AExpr\",     \"e\": %s }" (txt_of_aexpr e)
    | AReturn(e) -> sprintf "  { \"astmt\": \"AReturn \",     \"e\": %s }" (txt_of_aexpr e)
    | ABreak -> sprintf "  { \"astmt\": \"Break \"}"
    | AContinue -> sprintf "  { \"astmt\": \"Break \"}"
    | AWhile(e, s) -> sprintf "  { \"aexpr\": \"AWhile\",     \"e\": %s,     \"s\": %s }"  (txt_of_aexpr e) (txt_of_astmt s)
    | AForIn(e1, e2, s) -> sprintf "  { \"aexpr\": \"AForIn\",     \"e1\": %s,     \"e2\": %s,     \"s\": %s}"  (txt_of_aexpr e1)
    (txt_of_aexpr e2) (txt_of_astmt s)
    | AForRange(e1, e2, e3, s) -> sprintf "  { \"aexpr\": \"AForRange\",     \"e1\": %s,     \"e2\": %s,       \"e3\": %s,        \"s\": %s }" (txt_of_aexpr e1)
    (txt_of_aexpr e2) (txt_of_aexpr e3) (txt_of_astmt s)
    | AIf(e1, s1, s2) -> sprintf "  { \"aexpr\": \"AFIf\",     \"e1\": %s,     \"s1\": %s,     \"s2\": %s }" (txt_of_aexpr e1)

    (txt_of_astmt s1)  (txt_of_astmt s2)
    | ABlock(asl) -> (txt_of_astmts asl)
    | APrint(e) -> sprintf "  { \"astmt\": \"APrint\",     \"e\": %s }" (txt_of_aexpr e)
    | AObjectInit(obj_list) -> sprintf "  { \"astmt\": \"AObjectInit\",     \"objlist\": %s }" (txt_of_objlist obj_list)

and txt_of_astmts (astmts) =
  let rec aux acc = function
      | [] -> sprintf "[  %s  ] " (String.concat " , " (List.rev acc))
      | astmt :: tl -> aux (txt_of_astmt astmt :: acc) tl
  in aux [] astmts

(* Function declarations *)
and txt_of_afdecl (f: afunc_decl): string =
    sprintf " {    \"func name\" : \"%s\",        \"params\" : [%s],     \"body\": %s   }" (f.afname) 
			(txt_of_aformals f.aformals) (txt_of_astmts f.abody)


and txt_of_afdecls afdecls =
  let rec aux acc = function 
      | [] -> sprintf "[ %s ]" (String.concat " ,     " (List.rev acc))
      | afdecl :: tl -> aux ((txt_of_afdecl afdecl) :: acc) tl
  in aux [] afdecls

let print_sast aprogram =
    let sast_str = txt_of_afdecls aprogram in
    print_endline sast_str
